import unittest
import numpy as np

import peaksql.util


SEQ_LEN = 200


class TestUtil(unittest.TestCase):
    """ A test class to test the util of peaksql """

    def test_101_IUPAC_A_to_idx(self):
        onehot_idx = peaksql.util._nuc_to_onehot_idx.py_func(ord("A"))
        assert onehot_idx == 0

    def test_102_IUPAC_C_to_idx(self):
        onehot_idx = peaksql.util._nuc_to_onehot_idx.py_func(ord("C"))
        assert onehot_idx == 1

    def test_103_IUPAC_G_to_idx(self):
        onehot_idx = peaksql.util._nuc_to_onehot_idx.py_func(ord("G"))
        assert onehot_idx == 2

    def test_104_IUPAC_T_to_idx(self):
        onehot_idx = peaksql.util._nuc_to_onehot_idx.py_func(ord("T"))
        assert onehot_idx == 3

    def test_105_IUPAC_R_to_idx(self):
        onehot_idx = np.array(
            [peaksql.util._nuc_to_onehot_idx.py_func(ord("R")) for _ in range(SEQ_LEN)]
        )
        assert np.any(onehot_idx == 0)
        assert not np.any(onehot_idx == 1)
        assert np.any(onehot_idx == 2)
        assert not np.any(onehot_idx == 3)

    def test_106_IUPAC_Y_to_idx(self):
        onehot_idx = np.array(
            [peaksql.util._nuc_to_onehot_idx.py_func(ord("Y")) for _ in range(SEQ_LEN)]
        )
        assert not np.any(onehot_idx == 0)
        assert np.any(onehot_idx == 1)
        assert not np.any(onehot_idx == 2)
        assert np.any(onehot_idx == 3)

    def test_107_IUPAC_S_to_idx(self):
        onehot_idx = np.array(
            [peaksql.util._nuc_to_onehot_idx.py_func(ord("S")) for _ in range(SEQ_LEN)]
        )
        assert not np.any(onehot_idx == 0)
        assert np.any(onehot_idx == 1)
        assert np.any(onehot_idx == 2)
        assert not np.any(onehot_idx == 3)

    def test_108_IUPAC_W_to_idx(self):
        onehot_idx = np.array(
            [peaksql.util._nuc_to_onehot_idx.py_func(ord("W")) for _ in range(SEQ_LEN)]
        )
        assert np.any(onehot_idx == 0)
        assert not np.any(onehot_idx == 1)
        assert not np.any(onehot_idx == 2)
        assert np.any(onehot_idx == 3)

    def test_109_IUPAC_K_to_idx(self):
        onehot_idx = np.array(
            [peaksql.util._nuc_to_onehot_idx.py_func(ord("K")) for _ in range(SEQ_LEN)]
        )
        assert not np.any(onehot_idx == 0)
        assert not np.any(onehot_idx == 1)
        assert np.any(onehot_idx == 2)
        assert np.any(onehot_idx == 3)

    def test_110_IUPAC_M_to_idx(self):
        onehot_idx = np.array(
            [peaksql.util._nuc_to_onehot_idx.py_func(ord("M")) for _ in range(SEQ_LEN)]
        )
        assert np.any(onehot_idx == 0)
        assert np.any(onehot_idx == 1)
        assert not np.any(onehot_idx == 2)
        assert not np.any(onehot_idx == 3)

    def test_111_IUPAC_B_to_idx(self):
        onehot_idx = np.array(
            [peaksql.util._nuc_to_onehot_idx.py_func(ord("B")) for _ in range(SEQ_LEN)]
        )
        assert not np.any(onehot_idx == 0)
        assert np.any(onehot_idx == 1)
        assert np.any(onehot_idx == 2)
        assert np.any(onehot_idx == 3)

    def test_112_IUPAC_D_to_idx(self):
        onehot_idx = np.array(
            [peaksql.util._nuc_to_onehot_idx.py_func(ord("D")) for _ in range(SEQ_LEN)]
        )
        assert np.any(onehot_idx == 0)
        assert not np.any(onehot_idx == 1)
        assert np.any(onehot_idx == 2)
        assert np.any(onehot_idx == 3)

    def test_113_IUPAC_H_to_idx(self):
        onehot_idx = np.array(
            [peaksql.util._nuc_to_onehot_idx.py_func(ord("H")) for _ in range(SEQ_LEN)]
        )
        assert np.any(onehot_idx == 0)
        assert np.any(onehot_idx == 1)
        assert not np.any(onehot_idx == 2)
        assert np.any(onehot_idx == 3)

    def test_114_IUPAC_V_to_idx(self):
        onehot_idx = np.array(
            [peaksql.util._nuc_to_onehot_idx.py_func(ord("V")) for _ in range(SEQ_LEN)]
        )
        assert np.any(onehot_idx == 0)
        assert np.any(onehot_idx == 1)
        assert np.any(onehot_idx == 2)
        assert not np.any(onehot_idx == 3)

    def test_115_IUPAC_N_to_idx(self):
        onehot_idx = np.array(
            [peaksql.util._nuc_to_onehot_idx.py_func(ord("N")) for _ in range(SEQ_LEN)]
        )
        assert np.any(onehot_idx == 0)
        assert np.any(onehot_idx == 1)
        assert np.any(onehot_idx == 2)
        assert np.any(onehot_idx == 3)

    def test_116_IUPAC_wrong(self):
        self.assertRaises(ValueError, peaksql.util._nuc_to_onehot_idx.py_func, ord("Q"))

    def test_117_sequence_to_onehot(self):
        """
        Test whether single-nucleotide IUPAC codes are converted correctly into one-hot
        encoding.
        """
        sequence = "ACGTACGT"
        onehot = peaksql.util.sequence_to_onehot(sequence)
        true = np.array(
            [
                [True, False, False, False],
                [False, True, False, False],
                [False, False, True, False],
                [False, False, False, True],
                [True, False, False, False],
                [False, True, False, False],
                [False, False, True, False],
                [False, False, False, True],
            ],
            dtype=bool,
        )
        np.testing.assert_array_equal(onehot, true)

    def test_118_binary_search_out_of_bound(self):
        haystack = np.array([0, 5, 10, 12, 22])

        # call each function twice, once jit compiled and once not compiled for coverage
        self.assertRaises(ValueError, peaksql.util.binary_search, -1, haystack)
        self.assertRaises(ValueError, peaksql.util.binary_search.py_func, -1, haystack)

        self.assertRaises(ValueError, peaksql.util.binary_search, 22, haystack)
        self.assertRaises(ValueError, peaksql.util.binary_search.py_func, 22, haystack)

    def test_119_binary_search(self):
        haystack = np.array([0, 5, 10, 12, 22])

        # call each function twice, once jit compiled and once not compiled for coverage
        assert peaksql.util.binary_search(0, haystack) == 1
        assert peaksql.util.binary_search.py_func(0, haystack) == 1

        assert peaksql.util.binary_search(1, haystack) == 1
        assert peaksql.util.binary_search.py_func(1, haystack) == 1

        assert peaksql.util.binary_search(14, haystack) == 4
        assert peaksql.util.binary_search.py_func(14, haystack) == 4



"""
Collection of all the tables used by PeakSQL.
"""
# Assembly table
ASS = (
    "Assembly ("
    "    AssemblyId INTEGER PRIMARY KEY AUTOINCREMENT,"
    "    Assembly TEXT NOT NULL,"
    "    Species,"
    "    Size INT NOT NULL,"
    "    AbsPath TEXT UNIQUE NOT NULL"
    ")"
)

# Chromosome table
CHR = (
    "Chromosome ("
    "    ChromosomeId INTEGER PRIMARY KEY AUTOINCREMENT,"
    "    Chromosome TEXT,"
    "    Size INT NOT NULL,"
    "    Offset INT NOT NULL,"
    "    AssemblyId NOT NULL,"
    "    FOREIGN KEY(AssemblyId) REFERENCES Assembly(AssemblyId)"
    ")"
)

# Condition table
CON = (
    "Condition ("
    "    ConditionId INTEGER PRIMARY KEY AUTOINCREMENT,"
    "    Condition TEXT"
    ")"
)

# Bed table
BED = (
    "Bed ("
    "    BedId INTEGER PRIMARY KEY AUTOINCREMENT,"
    "    ConditionId,"
    "    ChromosomeId NOT NULL,"
    "    Peak INT,"  # narrowPeak summit
    "    FOREIGN KEY(ChromosomeId) REFERENCES Chromosome(ChromosomeId),"
    "    FOREIGN KEY(ConditionId)  REFERENCES Condition(ConditionId)"
    ")"
)

# Virtual Bed table, complement of the BED table. Uses r*tree for faster queries
BED_VIRT = (
    f"BedVirtual USING rtree("
    f"    BedId INT, "  # Foreign key not implemented :(
    f"    ChromStart INT, "
    f"    ChromEnd INT, "
    f")"
)



"""

"""
import sqlite3
import os
from functools import lru_cache

import pybedtools
import pyfaidx

import peaksql.tables as tables


class DataBase:
    """
    The DataBase class is an easy interface to store pre-processed NGS data for Machine
    Learning.

    It allows for ...
    """

    def __init__(self, db: str = "PeakSQL.sqlite", in_memory: bool = False):
        self.db = db

        # connect, and set a relatively high timeout number for multiprocessing
        self.conn = sqlite3.connect(db, timeout=30)
        self.cursor = self.conn.cursor()

        self.in_memory = in_memory
        if in_memory:
            # start a connection with our memory and move our database there
            dest = sqlite3.connect(":memory:")
            self.conn.backup(dest)

            # replace the old connection and cursor with our new in-memory connection
            self.conn = dest
            self.cursor = self.conn.cursor()

        # register all the tables (Assembly, Chromosome, Condition, Peak)
        for table in [table for table in dir(tables) if not table.startswith("__")]:
            virtual = "VIRTUAL" if "virtual" in getattr(tables, table).lower() else ""
            self.cursor.execute(
                f"CREATE {virtual} TABLE IF NOT EXISTS {getattr(tables, table)}"
            )

        self.conn.commit()

        # if we are loading a pre-existing database connect to all the assemblies
        self.cursor.execute("SELECT Assembly, AbsPath FROM Assembly")
        self.fastas = {
            assembly: pyfaidx.Fasta(abspath)
            for assembly, abspath in self.cursor.fetchall()
        }

    @lru_cache()
    def get_assembly_id(self, assembly_name: str) -> int:
        """
        Quickly get the AssemblyId based on Assembly (name).
        """
        result = self.cursor.execute(
            f"SELECT AssemblyId FROM Assembly "
            f"    WHERE Assembly='{assembly_name}' "
            f"LIMIT 1"
        ).fetchone()
        if result:
            return result[0]
        raise ValueError(f"Assembly {assembly_name} is not present in the database")

    @lru_cache(maxsize=2 ** 16)
    def get_chrom_id(self, assembly_id: int, chrom_name: str) -> int:
        """
        Quickly get the ChromosomeId based on assemblyId and Chromosome (name).
        """
        result = self.cursor.execute(
            f"SELECT ChromosomeId FROM Chromosome "
            f"    WHERE Chromosome='{chrom_name}' "
            f"    AND AssemblyId='{assembly_id}'"
            f"LIMIT 1"
        ).fetchone()
        if result:
            return result[0]
        raise ValueError(
            f"No chromosome {chrom_name} for assembly with assembly id {assembly_id}"
        )

    @property
    def assemblies(self):
        """
        return all registred assemblies
        """
        return [
            val[0]
            for val in self.cursor.execute(f"SELECT Assembly FROM Assembly").fetchall()
        ]

    def add_assembly(
        self, assembly_path: str, assembly: str = None, species: str = None
    ):
        """
        Add an assembly (genome) to the database. Sequences from the assembly are
        retrieved with PyFaidx, and the database assumes the assembly does not change
        location during during its lifetime.

        :param assembly_path: The path to the assembly file.
        :param assembly: The name of the assembly (optional: default is the name of the
                         file).
        :param species: The name of the species the assembly belongs to (optional:
                        default is the assembly name)
        """
        assert not self.in_memory, (
            "It is currently not supported to add data with an in-memory " "database."
        )
        # set defaults if none provided
        assembly = (
            assembly if assembly else os.path.basename(assembly_path).split(".")[0]
        )
        species = species if species else assembly
        abs_path = os.path.abspath(assembly_path)

        # TODO: what should default behaviour be (probably overwrite) or ignore??
        # TODO: check for assembly_path instead of name?
        # make sure the assembly hasn't been added yet
        assert (
            assembly not in self.assemblies
        ), f"Assembly '{assembly}' has already been added to the database!"

        fasta = pyfaidx.Fasta(abs_path)
        size = sum(len(seq) for seq in fasta.values())
        # add the assembly to the assembly table
        self.cursor.execute(
            f"INSERT INTO Assembly (Assembly, Species, Abspath, Size) "
            f"VALUES ('{assembly}', '{species}', '{abs_path}', {size})"
        )
        assembly_id = self.cursor.lastrowid

        # now fill the chromosome table
        offset = self.cursor.execute(
            f"SELECT SUM(Size) FROM Assembly WHERE AssemblyId < {assembly_id}"
        ).fetchone()[0]
        offset = 0 if offset is None else offset
        for sequence_name, sequence in fasta.items():
            size = len(sequence)
            self.cursor.execute(
                f"INSERT INTO Chromosome (AssemblyId, Size, Chromosome, Offset) "
                f"    VALUES('{assembly_id}', '{size}', '{sequence_name}', '{offset}')"
            )
            offset += size

        # clean up after yourself
        self.conn.commit()

    def add_data(self, data_path: str, assembly: str, condition: str = None):
        """
        Add data (bed or narrowPeak) to the database. The files are stored line by line

        :param data_path: The path to the assembly file.
        :param assembly: The name of the assembly. Requires the assembly to be added to
                         the database prior.
        :param condition: Experimental condition (optional). This allows for filtering
                          on conditions , e.g. when streaming data with a DataSet.
        """
        assert not self.in_memory, (
            "It is currently not supported to add data with an in-memory " "database."
        )
        # check for supported filetype
        *_, extension = os.path.splitext(data_path)
        # TODO: add more extensions
        assert extension in [
            ".narrowPeak",
            ".bed",
            ".bw",
        ], f"The file extension you choose is not supported"

        # check if species it belongs to has already been added to the database
        assembly_id = self.cursor.execute(
            f"SELECT AssemblyId FROM Assembly WHERE Assembly='{assembly}'"
        ).fetchone()
        assembly_id = assembly_id[0] if assembly_id else assembly_id
        assert assembly_id, (
            f"Assembly '{assembly}' has not been added to the database yet. Before "
            f"adding data you should add assemblies with the DataBase.add_assembly "
            f"method."
        )

        # Make sure that condition 'None' exists
        # This somehow locks the database when in __init__
        self.cursor.execute(
            "INSERT INTO Condition(ConditionId, Condition) SELECT 0, NULL "
            "WHERE NOT EXISTS(SELECT * FROM Condition WHERE ConditionId = 0)"
        )

        # get the condition id
        condition_id = (
            self.cursor.execute(
                f"SELECT ConditionId FROM Condition WHERE Condition='{condition}'"
            ).fetchone()
            if condition
            else (0,)
        )
        condition_id = condition_id[0] if condition_id else None

        # add the condition if necessary
        if condition and not condition_id:
            self.cursor.execute(f"INSERT INTO Condition VALUES(NULL, '{condition}')")

        if extension in [".bed", ".narrowPeak"]:
            self._add_bed(data_path, assembly_id, condition_id, extension)
        elif extension in [".bw"]:
            self._add_bigwig(data_path, assembly_id, condition_id, extension)

    def _add_bed(self, data_path, assembly_id, condition_id, extension):
        bed = pybedtools.BedTool(data_path)
        bed_lines = []
        virt_lines = []

        # get the current BedId we are at
        highest_id = self.cursor.execute(
            "SELECT BedId FROM Bed ORDER BY BedId DESC LIMIT 1"
        ).fetchone()
        if not highest_id:
            highest_id = 0
        else:
            highest_id = highest_id[0]

        for i, region in enumerate(bed):
            chromosome_id = self.get_chrom_id(assembly_id, region.chrom)
            offset = self.cursor.execute(
                f"SELECT Offset FROM Chromosome "
                f"WHERE ChromosomeId = {chromosome_id}"
            ).fetchone()[0]

            chromstart, chromend = region.fields[1:3]
            chromstart = int(chromstart) + offset
            chromend = int(chromend) + offset
            virt_lines.append((i + 1 + highest_id, chromstart, chromend))

            if len(region.fields) == 3 and extension == ".bed":
                bed_lines.append((condition_id, chromosome_id, None))
            elif len(region.fields) == 10 and extension == ".narrowPeak":
                bed_lines.append((condition_id, chromosome_id, region.fields[9]))
            else:
                fields = {".bed": 3, ".narrowPeak": 10}
                assert False, (
                    f"Extension {extension} should have {fields[extension]} fields, "
                    f"however it has {len(fields)}"
                )

        self.cursor.executemany(f"INSERT INTO Bed VALUES(NULL, ?, ?, ?)", bed_lines)

        # also add each bed entry to the BedVirtual table
        self.cursor.executemany(
            f"INSERT INTO BedVirtual VALUES(?, ?, ?)", virt_lines,
        )

        self.conn.commit()

    def _add_bigwig(self, data_path, assembly_id, condition_id, extension):
        raise NotImplementedError



import numpy as np
import multiprocessing
from abc import ABC, abstractmethod
from typing import Dict, Tuple

from ..database import DataBase
import peaksql.util as util


class _DataSet(ABC):
    """
    DataSet baseclass.
    """

    SELECT_CHROM_ASS = "SELECT Assembly, Chromosome "
    FROM = (
        " FROM Chromosome Chr "
        " INNER JOIN Assembly Ass  ON Chr.AssemblyId   = Ass.AssemblyId "
    )

    def __init__(self, database: str, where: str = "", seq_length: int = 200, **kwargs):
        # check for valid input
        if ("stride" in kwargs) == ("nr_rand_pos" in kwargs):  # xor
            raise ValueError("choose a stride OR a number of random positions")

        # store general stuff
        self.database_path = database
        self.databases: Dict[str, DataBase] = dict()
        self.seq_length = seq_length
        self.in_memory = kwargs.get("in_memory", False)

        # sql(ite) lookup
        self.WHERE = where
        query = (
            self.SELECT_CHROM_ASS + self.FROM + self.WHERE + " ORDER BY ChromosomeId"
        )
        self.database.cursor.execute(query)
        self.fetchall = self.database.cursor.fetchall()

        # get the genomic positions of our indices
        if "stride" in kwargs:
            self.stride = kwargs["stride"]
            self.chromosomes, self.cumsum, self.positions = self.get_strided_positions(
                self.seq_length, self.stride
            )
        if "nr_rand_pos" in kwargs:
            self.nr_rand_pos = kwargs["nr_rand_pos"]
            self.chromosomes, self.cumsum, self.positions = self.get_random_positions(
                self.seq_length, self.nr_rand_pos
            )

        # get all the conditions and their id in the database
        self.all_conditions = {
            k: v
            for k, v in self.database.cursor.execute(
                "SELECT DISTINCT Condition, ConditionId FROM Condition"
            ).fetchall()
        }

        # mark fetchall for garbage collection (large and we don't need it anymore)
        del self.fetchall

    def __len__(self) -> int:
        """
        Return the number of indices this dataset contains.
        """
        return self.cumsum[-1]

    def __getitem__(self, index: int) -> Tuple[np.array, int]:
        """
        Return the sequence in one-hot encoding and the label of the corresponding
        index.
        """
        if index >= len(self):
            raise StopIteration

        assembly, chrom, chromstart, chromend = self._index_to_site(index)

        # get the sequence, label and condition
        seq = self.get_onehot_sequence(assembly, chrom, chromstart, chromend)
        label = self.get_label(assembly, chrom, chromstart, chromend)

        return seq, label

    def _get_process(self) -> str:
        """
        PyFaidx is not multiprocessing safe when reading from fasta index or with
        sql(ite) queries. However if we start a new DataBase class for each process, we
        automatically start new sql(ite) connections and pyfaidx instances. This allows
        us to "stream" our data parallel in e.g. a Pytorch dataloader.
        """
        process = multiprocessing.current_process().name
        if process not in self.databases:
            self.databases[process] = DataBase(
                self.database_path, in_memory=self.in_memory
            )
        return process

    def _index_to_site(self, index: int) -> Tuple[str, str, int, int]:
        """
        Convert the index of self.__getitem__ to a tuple of (assembly, chrom,
        chromstart, chromend)

        Uses binary search for fast retrieval.
        """
        index_bs = util.binary_search(index, self.cumsum)

        assembly, chrom = self.chromosomes[index_bs]
        chromstart = self.positions[index_bs][index - self.cumsum[index_bs - 1]]
        chromend = chromstart + self.seq_length

        return assembly, chrom, chromstart, chromend

    def get_strided_positions(
        self, seq_length: int, stride: int
    ) -> Tuple[list, np.ndarray, np.ndarray]:
        """
        Calculate a map that connects __getitem__ indices to (assembly, chrom,
        chromstart) triplet. The positions are sampled accross the query with an even
        stride.

        The first return value is a list of (assembly, chrom) pairs, the second list
        consists of the cumulative sum of the number of indices that belong to this
        assembly, chrom pair, and the third list contains all chromstarts of the
        sequences. This allows for a decently fast and memory-efficient lookup of
        genomic positions corresponding to an index.
        """
        combis = [(None, None)] + list(
            {(assembly, chrom) for assembly, chrom, *_ in self.fetchall}
        )

        counts = [0]
        startpos = [np.array([])]
        non_empty_combis = [(None, None)]
        for assembly, chrom in combis[1:]:
            positions = np.arange(
                0, len(self.database.fastas[assembly][chrom]) - seq_length + 1, stride
            )
            if len(positions):
                non_empty_combis.append((assembly, chrom))
                startpos.append(positions)
                counts.append(len(startpos[-1]))

        cumsum = np.cumsum(counts)

        return non_empty_combis, cumsum, startpos

    def get_random_positions(
        self, seq_length: int, nr_rand_pos: int
    ) -> Tuple[list, np.ndarray, np.ndarray]:
        """
        Calculate a map that connects __getitem__ indices to (assembly, chrom,
        chromstart) triplet. The positions are sampled accross the query randomly, but
        proportional to the size of each chromosome.

        The first return value is a list of (assembly, chrom) pairs, the second list
        consists of the cumulative sum of the number of indices that belong to this
        assembly, chrom pair, and the third list contains all chromstarts of the
        sequences. This allows for a decently fast and memory-efficient lookup of
        genomic positions corresponding to an index.
        """
        # FIXME: this implementation is very buggy, let's just raise error for now
        combis = list({(assembly, chrom) for assembly, chrom, *_ in self.fetchall})
        combis = [(None, None)] + [
            (assembly, chrom)
            for assembly, chrom, *_ in self.fetchall
            if len(self.database.fastas[assembly][chrom]) > seq_length
        ]

        # distribute the positions over the chromosomes
        sizes = []
        for assembly, chrom in combis[1:]:
            sizes.append(len(self.database.fastas[assembly][chrom]))

        distribution = np.random.choice(
            range(len(sizes)), size=nr_rand_pos, p=np.array(sizes) / np.sum(sizes)
        )
        vals, counts = np.unique(distribution, return_counts=True)

        # then distribute inside a chromosome
        total_counts = [0]
        startpos = [np.array([])]
        non_empty_combis = [(None, None)]
        for i, (assembly, chrom) in enumerate(combis[1:]):
            where = np.where(vals == i)
            if len(where[0]) > 0:
                total_counts.append(counts[where[0][0]])
                startpos.append(
                    np.random.randint(
                        0,
                        len(self.database.fastas[assembly][chrom]) - seq_length,
                        size=counts[where[0][0]],
                    )
                )
                non_empty_combis.append((assembly, chrom))

        cumsum = np.cumsum(total_counts)

        return non_empty_combis, cumsum, startpos

    def get_onehot_sequence(
        self, assembly: str, chrom: str, chromstart: int, chromend: int
    ) -> np.ndarray:
        """
        Get the one-hot encoded sequence based on the assembly, chromosome, chromstart
        and chromend.
        """
        process = self._get_process()

        seq = self.databases[process].fastas[assembly][chrom][chromstart:chromend]
        seq = util.sequence_to_onehot(seq)

        return seq

    @abstractmethod
    def get_label(
        self, assembly: str, chrom: str, chromstart: int, chromend: int
    ) -> np.ndarray:
        pass

    @property
    def database(self):
        process = self._get_process()
        return self.databases[process]



import unittest
import sys
import os

import peaksql


DATABASE_BED = "test_peaksql_bed.sqlite"
DATABASE_NWP = "test_peaksql_narrowpeak.sqlite"

# make sure we begin with clean environment
if os.path.isfile(DATABASE_BED):
    os.remove(DATABASE_BED)
if os.path.isfile(DATABASE_NWP):
    os.remove(DATABASE_NWP)


class TestDataBase(unittest.TestCase):
    """ A test class to test the peaksql.DATABASE_BED """

    def setUp(self):
        """
        Due to Unit Test architecture it will give us lots of ResourceWarnings which
        aren't relevant to our tests. We turn them off.
        """
        if not sys.warnoptions:
            import warnings

            warnings.simplefilter("ignore")

    def test_201_load_genomes(self):
        db = peaksql.DataBase(DATABASE_BED)
        db.add_assembly("test/data/assembly1.fa")
        db.add_assembly("test/data/assembly2.fa")
        self.assertEquals(db.assemblies, ["assembly1", "assembly2"])
        self.assertRaises(AssertionError, db.add_assembly, "test/data/assembly1.fa")

    def test_202_test_get_assembly_id(self):
        db = peaksql.DataBase(DATABASE_BED)
        self.assertEquals(db.get_assembly_id("assembly1"), 1)
        self.assertEquals(db.get_assembly_id("assembly2"), 2)
        self.assertEquals(db.get_assembly_id("assembly2"), 2)
        self.assertRaises(ValueError, db.get_assembly_id, "assembly3")

    def test_203_get_chrom_id(self):
        db = peaksql.DataBase(DATABASE_BED)
        self.assertEquals(db.get_chrom_id(1, "chr1"), 1)
        self.assertEquals(db.get_chrom_id(1, "chr2"), 2)
        self.assertEquals(db.get_chrom_id(2, "chr1"), 3)
        self.assertEquals(db.get_chrom_id(2, "chr3"), 4)
        self.assertRaises(ValueError, db.get_chrom_id, 1, 3)

    def test_204_add_bed(self):
        db = peaksql.DataBase(DATABASE_BED)
        db.add_data("test/data/assembly1.bed", "assembly1")
        db.add_data("test/data/assembly2.bed", "assembly2")
        assert db.cursor.execute("SELECT COUNT(BedId) FROM Bed").fetchone() == (3,)
        assert db.cursor.execute("SELECT COUNT(BedId) FROM BedVirtual").fetchone() == (
            3,
        )
        assert db.cursor.execute(
            "SELECT Chromosome, ChromStart - Offset, ChromEnd - Offset "
            "FROM BED B "
            "JOIN Chromosome C on B.ChromosomeId = C.ChromosomeId "
            "JOIN BedVirtual BV on BV.BedId = B.BedId "
            "JOIN Assembly A on C.AssemblyId = A.AssemblyId "
            "WHERE Assembly='assembly2' and Chromosome='chr1'"
        ).fetchall() == [("chr1", 20, 40)]

    def test_205_add_narrowpeak(self):
        db = peaksql.DataBase(DATABASE_NWP)
        db.add_assembly("test/data/assembly1.fa")
        db.add_data("test/data/assembly1.narrowPeak", "assembly1")
        assert db.cursor.execute("SELECT COUNT(BedId) FROM Bed").fetchone() == (4,)
        assert db.cursor.execute("SELECT COUNT(BedId) FROM BedVirtual").fetchone() == (
            4,
        )
        assert db.cursor.execute(
            "SELECT Chromosome, ChromStart - Offset, ChromEnd - Offset "
            "FROM BED B "
            "JOIN Chromosome C on B.ChromosomeId = C.ChromosomeId "
            "JOIN BedVirtual BV on BV.BedId = B.BedId "
            "JOIN Assembly A on C.AssemblyId = A.AssemblyId "
            "WHERE Assembly='assembly1' and Chromosome='chr1'"
        ).fetchall() == [("chr1", 0, 10), ("chr1", 20, 30)]

    def test_206_in_memory(self):
        db_file = peaksql.DataBase(DATABASE_BED, in_memory=False)
        db_memo = peaksql.DataBase(DATABASE_BED, in_memory=True)

        assert (
            db_file.cursor.execute("SELECT * FROM Assembly").fetchall()
            == db_memo.cursor.execute("SELECT * FROM Assembly").fetchall()
        )
        assert (
            db_file.cursor.execute("SELECT * FROM Chromosome").fetchall()
            == db_memo.cursor.execute("SELECT * FROM Chromosome").fetchall()
        )
        assert (
            db_file.cursor.execute("SELECT * FROM BED").fetchall()
            == db_memo.cursor.execute("SELECT * FROM BED").fetchall()
        )



__version__ = "0.0.0"

# make sure that relevant stuff is importable
from . import database, util
from .database import DataBase
from .datasets.narrowpeak import NarrowPeakDataSet
from .datasets.bedregion import BedRegionDataSet

__all__ = ["database", "util", "DataBase", "NarrowPeakDataSet", "BedRegionDataSet"]



# TODO: random seed + tests
import numba
import numpy as np


@numba.jit(nopython=True, cache=True)
def _nuc_to_onehot_idx(nuc: int) -> int:
    """
    Convert a nucleotide to a one hot index, where the indexes 0, 1, 2, 3 respectively
    correspond to A, C, G, T.

    Accepts all IUPAC nucleotide codes, and picks a random option from the possible
    nucleotides.
    """
    # first try single-nucleotide codes
    # start with A & T, since genomes generally have higher AT content than GC content
    if nuc == 65:  # A
        return 0
    if nuc == 84:  # T
        return 3
    if nuc == 67:  # C
        return 1
    if nuc == 71:  # G
        return 2

    # if that doesn't work try multiple-nucleotide codes
    # first set potential indexes
    if nuc == 78:  # N
        idx = np.array([0, 1, 2, 3])
    elif nuc == 82:  # R
        idx = np.array([0, 2])
    elif nuc == 89:  # Y
        idx = np.array([1, 3])
    elif nuc == 83:  # S
        idx = np.array([1, 2])
    elif nuc == 87:  # W
        idx = np.array([0, 3])
    elif nuc == 75:  # K
        idx = np.array([2, 3])
    elif nuc == 77:  # M
        idx = np.array([0, 1])
    elif nuc == 66:  # B
        idx = np.array([1, 2, 3])
    elif nuc == 68:  # D
        idx = np.array([0, 2, 3])
    elif nuc == 72:  # H
        idx = np.array([0, 1, 3])
    elif nuc == 86:  # V
        idx = np.array([0, 1, 2])
    else:
        raise ValueError("Only IUPAC nucleotide codes are accepted.")

    # now return any of the possible indexes
    return np.random.choice(idx)


@numba.jit(nopython=True, cache=True)
def _sequence_to_onehot(sequence: np.ndarray) -> np.ndarray:
    onehot = np.zeros((len(sequence), 4), dtype=numba.boolean)
    for i, nuc in enumerate(sequence):
        onehot[i, _nuc_to_onehot_idx(nuc)] = True

    return onehot


def sequence_to_onehot(sequence) -> np.ndarray:
    """
    Convert a sequence of length n to one-hot encoding of shape (n x 4).
    """
    return _sequence_to_onehot(str(sequence).upper().encode("utf-8"))


@numba.jit(nopython=True, cache=True)
def binary_search(index: int, lens: np.ndarray) -> int:
    """
    Does a binary search to not find the value in a list, but the index where the value
    is in between two values (returns the higher index of the two).

    index: 5
    lens: [0, 4, 6, 8, 22]
    returns: 2
    """
    if index < lens[0] or index >= lens[-1]:
        raise ValueError("Invalid index")

    # FIXME: this can hang in infinite loop, and because of numba becomes unresponsive.
    left, right = 0, len(lens)

    while left <= right:
        mid = (left + right) // 2

        if lens[mid - 1] > index:
            right = mid

        elif lens[mid] <= index:
            left = mid

        else:
            return mid

    assert False



import numpy as np
from abc import ABC, abstractmethod
from typing import List, Tuple

from .base import _DataSet


class _BedDataSet(_DataSet, ABC):
    """
    The BedDataSet...
    """

    FROM = (
        " FROM Chromosome Chr "
        " INNER JOIN Assembly Ass  ON Chr.AssemblyId   = Ass.AssemblyId "
    )
    SELECT_LABEL: str

    def __init__(self, database: str, where: str = "", seq_length: int = 200, **kwargs):
        _DataSet.__init__(self, database, where, seq_length, **kwargs)

        assert "label_func" in kwargs and kwargs["label_func"] in [
            "any",
            "inner_any",
            "all",
            "inner_all",
            "fraction",
            "inner_fraction",
        ]

        if "inner" in kwargs["label_func"]:
            if "inner_range" not in kwargs:
                raise ValueError(
                    f"You specified an 'inner' function {kwargs['label_func']} but did "
                    f"not specify an inner_range."
                )
            self.inner_range = kwargs["inner_range"]

        if "fraction" in kwargs["label_func"]:
            if "fraction" not in kwargs:
                raise ValueError(
                    f"You specified a 'fraction' of the sequence to be within a ragion,"
                    f" but you did not not specify the fraction."
                )
            self.fraction = kwargs["fraction"]

        setattr(self, "label_from_array", eval("self.label_" + kwargs["label_func"]))

    def label_any(self, positions: np.ndarray) -> np.ndarray:
        return np.any(positions, axis=1)

    def label_inner_any(self, positions: np.ndarray) -> np.ndarray:
        mid = positions.shape[0] // 2
        return self.label_any(
            positions[:, mid - self.inner_range : mid + self.inner_range + 1]
        )

    def label_all(self, positions: np.ndarray) -> np.ndarray:
        return np.all(positions, axis=1)

    def label_inner_all(self, positions: np.ndarray) -> np.ndarray:
        mid = positions.shape[0] // 2
        return self.label_all(
            positions[:, mid - self.inner_range : mid + self.inner_range + 1]
        )

    def label_fraction(self, positions: np.ndarray) -> np.ndarray:
        return np.sum(positions, axis=1) / positions.shape[1] >= self.fraction

    def label_inner_fraction(self, positions: np.ndarray) -> np.ndarray:
        mid = positions.shape[0] // 2
        return self.label_fraction(
            positions[:, mid - self.inner_range : mid + self.inner_range + 1]
        )

    @abstractmethod
    def array_from_query(
        self,
        query: List[Tuple[int, int, int, int]],
        cur_chrom_id: int,
        chromstart: int,
        chromend: int,
    ) -> np.ndarray:
        pass

    def label_from_array(self, positions: np.ndarray) -> np.ndarray:
        raise NotImplementedError

    def get_label(
        self, assembly: str, chrom: str, chromstart: int, chromend: int
    ) -> np.ndarray:
        """
        Get the label that corresponds to chromstart:chromend.
        """
        assemblyid = self.database.get_assembly_id(assembly)
        chromosomeid = self.database.get_chrom_id(assemblyid, chrom)

        offset = self.database.cursor.execute(
            f"""
            SELECT Offset FROM Chromosome WHERE ChromosomeId = {chromosomeid}
            """
        ).fetchone()[0]
        chromstart += offset
        chromend += offset

        query = f"""
            SELECT {self.SELECT_LABEL}
            FROM BedVirtual
            INNER JOIN Bed on BedVirtual.BedId = Bed.BedId
            WHERE ({chromstart} <= BedVirtual.ChromEnd) AND
                  ({chromend} >= BedVirtual.ChromStart)
        """
        query_result = self.database.cursor.execute(query).fetchall()
        positions = self.array_from_query(
            query_result, chromosomeid, chromstart, chromend
        )
        labels = self.label_from_array(positions)

        return labels



from setuptools import setup
import toml

project = toml.load("pyproject.toml")["project"]

# read the readme as long description
with open("README.md") as f:
    project["long_description"] = f.read()

project["long_description_content_type"] = "text/markdown"

setup(**project)



import numpy as np
from abc import ABC
from typing import List, Tuple

from .bed import _BedDataSet


class BedRegionDataSet(_BedDataSet, ABC):
    """
    The BedRegion ...
    """

    SELECT_LABEL = (
        " Bed.ChromosomeId, Bed.ConditionId, BedVirtual.ChromStart, BedVirtual.ChromEnd"
    )

    def __init__(
        self,
        database: str,
        where: str = "",
        seq_length: int = 200,
        label_func: str = "any",
        **kwargs
    ):
        _BedDataSet.__init__(
            self, database, where, seq_length, label_func=label_func, **kwargs
        )

    def array_from_query(
        self,
        query: List[Tuple[int, int, int, int]],
        cur_chrom_id: int,
        chromstart: int,
        chromend: int,
    ) -> np.ndarray:
        positions = np.zeros(
            (len(self.all_conditions), chromend - chromstart), dtype=bool
        )

        for chromosome_id, condition_id, start, end in query:
            if cur_chrom_id == chromosome_id:
                min_idx = max(0, int(start - chromstart))
                max_idx = min(positions.shape[1], int(end - chromstart))

                positions[condition_id, min_idx:max_idx] = True

        return positions



import unittest
import numpy as np

import peaksql
from test.test_02_database import DATABASE_BED, DATABASE_NWP


class TestDataBase(unittest.TestCase):
    """ A test class to test the peaksql.datasets._BedDataSet family """

    positions = np.array(
        [
            [False, True, False, False, False],
            [False, True, True, False, False],
            [False, False, False, False, False],
            [True, True, True, True, True],
        ]
    )

    def test_301_BedRegionDataSet_stride_length(self):
        dataset = peaksql.BedRegionDataSet(DATABASE_BED, seq_length=10, stride=10)
        assert len(dataset) == 16

    def test_302_BedRegionDataSet_stride_sequences(self):
        dataset = peaksql.BedRegionDataSet(DATABASE_BED, seq_length=10, stride=10)

        all_dna = (
            "AAAACCCCGGGGTTTTAAACCCGGGTTTAACCGGTTACGT"
            + "TTTTGGGGCCCCAAAATTTGGGCCCAAATTGGCCAATGCA"
            + "ATGCGTAGCTGATCGATGCTAGCTAGCTAGCTAGCTAAAA"
            + "ATGGTGAATGTGAGTAGTGATGATGAGTGTAGTGAGGGGG"
        )

        dna_strided = [all_dna[i : i + 10] for i in range(0, len(all_dna), 10)]
        dna_onehot = [peaksql.util.sequence_to_onehot(dna) for dna in dna_strided]
        for seq, label in dataset:
            assert np.sum(np.all(seq == potential_seq) for potential_seq in dna_onehot)

    def test_303_Bed_label_any(self):
        dataset = peaksql.BedRegionDataSet(DATABASE_BED, seq_length=10, stride=10)
        assert all(dataset.label_any(self.positions) == [True, True, False, True])

    def test_304_Bed_label_all(self):
        dataset = peaksql.BedRegionDataSet(DATABASE_BED, seq_length=10, stride=10)
        assert all(dataset.label_all(self.positions) == [False, False, False, True])

    def test_305_Bed_label_fraction(self):
        dataset = peaksql.BedRegionDataSet(DATABASE_BED, seq_length=10, stride=10)
        dataset.fraction = 0.4
        assert all(dataset.label_fraction(self.positions) == [False, True, False, True])

    def test_306_Bed_label_inner_any(self):
        dataset = peaksql.BedRegionDataSet(DATABASE_BED, seq_length=10, stride=10)
        dataset.inner_range = 1
        assert all(dataset.label_any(self.positions) == [True, True, False, True])

    def test_307_Bed_label_inner_all(self):
        dataset = peaksql.BedRegionDataSet(DATABASE_BED, seq_length=10, stride=10)
        dataset.inner_range = 1
        assert all(
            dataset.label_inner_all(self.positions) == [False, False, False, True]
        )

    def test_308_Bed_label_inner_fraction(self):
        dataset = peaksql.BedRegionDataSet(DATABASE_BED, seq_length=10, stride=10)
        dataset.inner_range = 1
        dataset.fraction = 2 / 3
        assert all(
            dataset.label_inner_fraction(self.positions) == [False, True, False, True]
        )

    def test_309_BedRegionDataSet_array_from_query(self):
        chromstart = 10
        chromend = 20
        chromid = 1
        query = [[0, 0, 15, 25], [1, 0, 5, 13]]
        dataset = peaksql.BedRegionDataSet(DATABASE_BED, seq_length=10, stride=10)
        assert np.all(
            dataset.array_from_query(query, chromid, chromstart, chromend)
            == np.array(
                [[True, True, True, False, False, False, False, False, False, False]]
            )
        )

    def test_310_NarrowPeak_array_from_query(self):
        chromstart = 10
        chromend = 20
        chromid = 1
        query = [[0, 0, 15, 25], [1, 0, 5, 13]]
        dataset = peaksql.NarrowPeakDataSet(DATABASE_NWP, seq_length=10, stride=10)
        assert np.all(
            dataset.array_from_query(query, chromid, chromstart, chromend)
            == np.array(
                [[False, False, False, False, False, False, False, False, True, False]]
            )
        )

    def test_311_BedRegionDataSet_random_pos_length(self):
        dataset = peaksql.BedRegionDataSet(DATABASE_BED, seq_length=10, nr_rand_pos=20)
        assert len(dataset) == 20

    def test_312_BedRegionDataSet_random_pos_sequences(self):
        dataset = peaksql.BedRegionDataSet(DATABASE_BED, seq_length=10, nr_rand_pos=20)
        all_dna = [
            "AAAACCCCGGGGTTTTAAACCCGGGTTTAACCGGTTACGT",
            "TTTTGGGGCCCCAAAATTTGGGCCCAAATTGGCCAATGCA",
            "ATGCGTAGCTGATCGATGCTAGCTAGCTAGCTAGCTAAAA",
            "ATGGTGAATGTGAGTAGTGATGATGAGTGTAGTGAGGGGG",
        ]
        dna_onehot = [peaksql.util.sequence_to_onehot(dna) for dna in all_dna]
        for i, (seq, label) in enumerate(dataset):
            found = False
            for chromosome in range(4):
                for idx in range(0, 30):
                    if np.all(seq == dna_onehot[chromosome][idx : idx + 10]):
                        found = True
            assert found

    def test_313_BedRegionDataSet_random_pos_distribution(self):
        dataset = peaksql.BedRegionDataSet(
            DATABASE_BED, seq_length=10, nr_rand_pos=100_000
        )

        # chromosomes are of equal size, so we expect equal nr of positions for each
        un_cumsum = dataset.cumsum - np.roll(dataset.cumsum, shift=1)
        for count in un_cumsum[1:]:
            assert 0.245 <= count / 100_000 <= 0.255

    def test_314_BedDataSet_label_func_parsing(self):
        with self.assertRaises(ValueError):
            peaksql.BedRegionDataSet(
                DATABASE_BED, seq_length=10, nr_rand_pos=10, label_func="inner_any"
            )
        with self.assertRaises(ValueError):
            peaksql.BedRegionDataSet(
                DATABASE_BED, seq_length=10, nr_rand_pos=10, label_func="inner_all"
            )
        with self.assertRaises(ValueError):
            peaksql.BedRegionDataSet(
                DATABASE_BED,
                seq_length=10,
                nr_rand_pos=10,
                label_func="inner_fraction",
                fraction=0.8,
            )
        with self.assertRaises(ValueError):
            peaksql.BedRegionDataSet(
                DATABASE_BED, seq_length=10, nr_rand_pos=10, label_func="fraction"
            )
